/**
  *  \file lfile.cc
  */

#include <vector>
#include <iostream>
#include "lfile.h"

/** \class LFile
    \brief Lisp File

    This is a file as accessible to the Lisp program. */
LFile::LFile(FILE* file, bool reading)
    : file(file), reading(reading)
{
    if (!file)
        throw LLispException("open-error", "unable to open file");
}

LFile::~LFile()
{
    close();
}

void
LFile::close()
{
    if (file)
        fclose(file);
    file = 0;
}

std::string
LFile::toString()
{
    if (!file)
        return "#<closed-stream>";
    else if (reading)
        return "#<input-stream>";
    else
        return "#<output-stream>";
}

bool
LFile::equals(LPtr other, bool /*identical*/)
{
    return other == this;
}

/******************************* Functions *******************************/

LFile*
argFile(LPtr& p)
{
    LPtr a = getArg(p);
    LFile* f = a.as<LFile>();
    if (!f)
        throw LLispException("type-error", "expected file");
    if (!f->getFile())
        throw LLispException("file-closed", "file already closed");
    return f;
}

//- (file-open-read FILE [BINFLAG])
//- => Open FILE for reading. If BINFLAG is non-nil, open in binary mode,
//-    otherwise open in text mode. Returns the stream object.
static LPtr
FFileOpenRead(LPtr p)
{
    std::string name = argString(p);
    bool bin = p.nonnull() ? argBool(p) : false;
    return new LFile(fopen(name.c_str(), bin ? "rb" : "r"), true);
}

//- (file-open-write FILE [BINFLAG])
//- => Create FILE, and open it for writing. If BINFLAG is non-nil, open
//-    in binary mode, otherwise open in text mode. Returns the stream
//-    object.
static LPtr
FFileOpenWrite(LPtr p)
{
    std::string name = argString(p);
    bool bin = p.nonnull() ? argBool(p) : false;
    return new LFile(fopen(name.c_str(), bin ? "wb" : "w"), false);
}

//- (file-write STREAM STRING....)
//- => Write all STRINGs to STREAM.
static LPtr
FFileWrite(LPtr p)
{
    LFile* f = argFile(p);
    while (p.nonnull()) {
        std::string s = argString(p);
        if (!f->isWriting())
            throw LLispException("file-mode", "file not open for writing");
        if (fwrite(s.data(), 1, s.size(), f->getFile()) != s.size())
            throw LLispException("io-error", "write error");
    }
    return f;
}

//- (file-read STREAM [LENGTH])
//- => Read from STREAM. If LENGTH is specified, read that many characters
//-    and throw io-error if there are not enough. Returns a string consisting
//-    of these characters.
//-    If LENGTH is omitted, reads a line. Returns the line, or nil on EOF.
static LPtr
FFileRead(LPtr p)
{
    LFile* f = argFile(p);
    if (p.nonnull()) {
        // read bytes
        int n = argInt(p);
        if (n < 0)
            throw LLispException("range-error", "size out of range");
        if (!f->isReading())
            throw LLispException("file-mode", "file not open for reading");
        if (n != 0) {
            std::vector<char> buf(n);
            if (fread(&buf.front(), 1, n, f->getFile()) != n)
                throw LLispException("io-error", "read error");
            return new LString(std::string(&buf.front(), buf.size()));
        } else {
            return new LString(std::string());
        }
    } else {
        // read line
        std::string s;
        int n;
        if (!f->isReading())
            throw LLispException("file-mode", "file not open for reading");
        while ((n = fgetc(f->getFile())) != EOF && n != '\n')
            s.append(1, char(n));
        if (s.size() == 0 && n != '\n')
            return LPtr();
        else
            return new LString(s);
    }
}

//- (file-set-pos STREAM POS)
//- => Change the file position, that is, the position where the next
//-    read or write operation will happen.
static LPtr
FFileSetPos(LPtr p)
{
    LFile* f = argFile(p);
    int n = argInt(p);
    fseek(f->getFile(), n, SEEK_SET);
    return LPtr();
}

//- (file-get-pos STREAM)
//- => Returns the current file position.
static LPtr
FFileGetPos(LPtr p)
{
    LFile* f = argFile(p);
    return new LInt(ftell(f->getFile()));
}

//- (file-close STREAM)
//- => Closes the stream. This will render it unusable. This call is not
//-    strictly necessary, the stream will automatically be closed when
//-    it is deallocated.
static LPtr
FFileClose(LPtr p)
{
    LFile* f = argFile(p);
    f->close();
    return LPtr();
}

//- (pack FORMAT ARGS....)
//- => A stripped-down version of Perl's 'pack'. FORMAT is a string specifying
//-    how to pack, ARGS are the values to pack. Returns the result string.
//-    The FORMAT contains type letters, optionally followed by a decimal
//-    COUNT or a '*'. Except for a and A, the COUNT means to convert that
//-    many arguments, * means to convert all remaining arguments (i.e. it must
//-    be the last character of FORMAT).
//-      a         argument is a string, pad it to COUNT bytes with null
//-                bytes; truncate it if it is too long.
//-      A         like a, but pad with spaces.
//-      c, C      argument is an integer, yields one character.
//-      s, S      argument is an integer, yields two characters containing
//-                the integer as 16-bit little-endian short.
//-      l,L,v,V   argument is an integer, yields four characters containing
//-                the integer as 32-bit little-endian long.
static LPtr
FPack(LPtr p)
{
    std::string s = argString(p);
    std::string::size_type n = 0;
    std::string accum;
    while (n < s.size()) {
        // parse
        char c = s[n++];
        int count = 1;
        if (n < s.size()) {
            if (s[n] >= '0' && s[n] <= '9') {
                count = 0;
                while (n < s.size() && s[n] >= '0' && s[n] <= '9') {
                    count = 10*count + s[n++] - '0';
                    if (count < 0)
                        throw LLispException("range-error", "numeric overflow in 'pack'");
                }
            } else if (s[n] == '*') {
                count = -1;
                ++n;
            }
        }
        // now interpret it
        switch (c) {
         case 'a': {
            // string, null-padded
            std::string s = argString(p);
            if (count > 0) {
                if (s.size() < count)
                    s.append(count - s.size(), char(0));
                else if (s.size() > count)
                    s.erase(count);
            }
            accum += s;
            count = 0;
            break;
         }
         case 'A': {
            // string, space-padded
            std::string s = argString(p);
            if (count > 0) {
                if (s.size() < count)
                    s.append(count - s.size(), ' ');
                else if (s.size() > count)
                    s.erase(count);
            }
            accum += s;
            count = 0;
            break;
         }
         case 'c':
         case 'C':
            // signed/unsigned char
            while (count != 0 && p.nonnull()) {
                accum += char(argInt(p));
                if (count > 0)
                    --count;
            }
            break;
         case 's':
         case 'S':
            // signed short
            // unsigned short
            while (count != 0 && p.nonnull()) {
                unsigned short s = argInt(p);
                accum += char(s & 255);
                accum += char(s >> 8);
                if (count > 0)
                    --count;
            }
            break;
         case 'l': case 'v':
         case 'L': case 'V':
            // signed int
            // unsigned int
            while (count != 0 && p.nonnull()) {
                unsigned long s = argInt(p);
                accum += char(s & 255);
                accum += char(s >> 8);
                accum += char(s >> 16);
                accum += char(s >> 24);
                if (count > 0)
                    --count;
            }
            break;
         default:
            throw LLispException("range-error", "invalid type code in 'pack'");
        }
        if (count > 0)
            throw LLispException("format-args", "too few arguments for format in 'pack'");
    }
    return new LString(accum);
}

//- (unpack FORMAT STRING [POS])
//- => A stripped-down version of Perl's 'unpack'. FORMAT is a string specifying
//-    how to unpack, and can contain only one format. Unpacks the beginning of
//-    STRING or, if POS is given, from that position onwards. Like for pack,
//-    the FORMAT contains a type letter, optionally followed by a decimal
//-    COUNT or a '*'. COUNT or * is only permitted for a/A.
//-      a         extracts up to COUNT bytes, remove trailing null bytes,
//-                returns a string. a* extracts all the rest of STRING.
//-      A         like a, but remove trailing spaces.
//-      c         extract one character, interpret it as signed 8-bit, and
//-                return as integer.
//-      C         like c, but unsigned 8-bit.
//-      s, S      extract two characters, interpret as signed/unsigned 16-bit
//-                little-endian integer, and return it.
//-      l,L,v,V   extract four characters, interpret as signed/unsigned 32-bit
//-                little-endian integer, and return it.
static LPtr
FUnpack(LPtr p)
{
    std::string pat = argString(p);
    std::string src = argString(p);
    std::string::size_type start = p.nonnull() ? argInt(p) : 0;
    if (start > src.size())
        throw LLispException("range-error", "string index out of range in 'unpack'");
    if (pat.size() == 0)
        throw LLispException("range-error", "unpack pattern too short");
    std::string::size_type n = 0;
    char c = pat[n++];
    int count = 1;
    if (n < pat.size()) {
        if (pat[n] >= '0' && pat[n] <= '9') {
            count = 0;
            while (n < pat.size() && pat[n] >= '0' && pat[n] <= '9') {
                count = 10*count + pat[n++] - '0';
                if (count < 0)
                    throw LLispException("range-error", "numeric overflow in 'unpack'");
            }
        } else if (pat[n] == '*') {
            count = -1;
            ++n;
        }
    }
    if (n < pat.size())
        throw LLispException("range-error", "format in 'unpack' is to long");

    switch (c) {
     case 'a':
     case 'A': {
        // string, null-padded
        // string, space-padded
        std::string res;
        char rem;
        if (count < 0)
            res = src.substr(start);
        else
            res = src.substr(start, count);
        if (c == 'a')
            rem = 0;
        else
            rem = ' ';
        while (res.size() && res[res.size()-1] == rem)
            res.erase(res.size()-1);
        return new LString(res);
     }
     case 'c':
        if (start < src.size())
            return new LInt((signed char)(src[start]));
        else
            return LPtr();
     case 'C':
        if (start < src.size())
            return new LInt((unsigned char)(src[start]));
        else
            return LPtr();
     case 's':
     case 'S':
     case 'l': case 'v':
     case 'L': case 'V': {
        int n = (c == 's' || c == 'S' ? 2 : 4);
        unsigned long val = 0, scale = 1;
        if (start >= src.size())
            return LPtr();
        for (int i = 0; i < n; ++i) {
            if (start < src.size())
                val += scale * (unsigned char) src[start++];
            scale *= 256;
        }
        if (c == 's' && (val & 0x8000))
            return new LInt(val - 0x10000L);
        else
            return new LInt(val);
     }
     default:
        throw LLispException("range-error", "invalid type code in 'pack'");
    }
}

void
initFile()
{
    LSymbol::define("file-open-read",     new LBuiltin(FFileOpenRead,  1, 2));
    LSymbol::define("file-open-write",    new LBuiltin(FFileOpenWrite, 1, 2));
    LSymbol::define("file-write",         new LBuiltin(FFileWrite,     1, INT_MAX));
    LSymbol::define("file-read",          new LBuiltin(FFileRead,      1, 2));
    LSymbol::define("file-close",         new LBuiltin(FFileClose,     1, 1));
    LSymbol::define("file-set-pos",       new LBuiltin(FFileSetPos,    2, 2));
    LSymbol::define("file-get-pos",       new LBuiltin(FFileGetPos,    1, 1));
    LSymbol::define("pack",               new LBuiltin(FPack,          1, INT_MAX));
    LSymbol::define("unpack",             new LBuiltin(FUnpack,        2, 3));
}
